"""The `femesh` module handles mesh generation for planingFSI cases."""
from __future__ import annotations

import abc
import itertools
from collections.abc import Iterable
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np

from planingfsi import logger
from planingfsi import trig
from planingfsi.solver import fzero
from planingfsi.writers import write_as_list


class Mesh:
    """A high-level class to store a structural mesh.

    The mesh serves as a container for `Point` instances. It then serves as a container for
    any number of `Submesh` instances, which each consists of a series of `Curve`s and can
    optionally be associated with a planing surface.

    Attributes:
        points: A list of all points in the mesh.
        submesh: A list of submeshes belonging to the parent mesh.

    """

    def __init__(self) -> None:
        self.points: list["Point"] = []
        self.submesh: list["Subcomponent"] = []
        self.add_point(0, "dir", [0, 0])

    @property
    def curves(self) -> list["Curve"]:
        """A list of all component curves from all submeshes."""
        return list(itertools.chain(*(submesh.curves for submesh in self.submesh)))

    def get_point(self, pt_id: int, /) -> "Point":
        """Return a point by ID from the mesh."""
        # TODO: Potentially replace this with a dictionary lookup
        for pt in self.points:
            if pt.id == pt_id:
                return pt

        raise ValueError(f"Cannot find Point object with ID={pt_id}")

    """A method alias, kept for backwards compatibility with old meshDict files."""
    get_pt_by_id = get_point

    def add_submesh(self, name: str = "") -> "Subcomponent":
        """Add a submesh to the mesh."""
        submesh = Subcomponent(name, mesh=self)
        self.submesh.append(submesh)
        return submesh

    def add_point(
        self, id_: int, method: str, position: Iterable[float | int | str], **kwargs: Any
    ) -> "Point":
        """Add a new point to the mesh, returning the created point.

        Args:
            id_: The point ID.
            method: A method by which to specify the point. Depending on the method, the position
                argument is treated differently.

                Options are:
                    * "dir" - Direct coordinate specification
                    * "rel" - Relative to another point, using polar coordinates
                    * "con" - Constrained along a specific direction up to a certain distance or intersection
                    * "pct" - At a specific percentage of the distance between two other points
            position: An iterable of parameters, used for specifying the position.
                The specific meaning of the different parameters differs based on the selected method.
        Returns:
            The point that is created.

        """
        # TODO: Split method into several

        if method == "dir":
            # Direct coordinate specification
            point = Point(id=id_, mesh=self)
            self.points.append(point)
            point.position = np.array(position)
        elif method == "rel":
            # Relative coordinate specification using polar coordinates
            point = Point(id=id_, mesh=self)
            self.points.append(point)
            base_pt_id, ang, radius = position

            # TODO: Consider removing this and fixing static types
            assert isinstance(radius, (float, int)), type(radius)
            assert isinstance(ang, (float, int)), type(ang)

            point.position = self.get_point(int(base_pt_id)).position + radius * trig.angd2vec2d(
                float(ang)
            )
        elif method == "con":
            # Constrained coordinate specification
            # Either extrapolating at an angle or going horizontally or vertically
            point = Point(id=id_, mesh=self)
            self.points.append(point)
            base_pt_id, dim, val = position
            ang = kwargs.get("angle", 0.0 if dim == "x" else 90.0)

            base_pt = self.get_point(int(base_pt_id)).position
            if dim == "x":
                point.position = np.array(
                    [val, base_pt[1] + (val - base_pt[0]) * trig.tand(float(ang))]
                )
            elif dim == "y":
                point.position = np.array(
                    [base_pt[0] + (val - base_pt[1]) / trig.tand(float(ang)), val]
                )
            else:
                raise ValueError("Incorrect dimension specification")
        elif method == "pct":
            # Place a point at a certain percentage along the line between two other points
            base_pt_id, end_pt_id, pct = position
            curve = Curve(
                self.get_point(int(base_pt_id)),
                self.get_point(int(end_pt_id)),
            )
            point = self.add_point_along_curve(id_, curve=curve, pct=float(pct))
        else:
            raise NameError(f"Incorrect position specification method for point, ID: {id_}")

        return point

    def add_point_along_curve(self, id_: int, curve: "Curve", pct: float) -> "Point":
        """Add a point at a certain percentage along a curve."""
        point = Point(id=id_, mesh=self)
        self.points.append(point)
        point.position = curve.get_coords(pct)
        return point

    def add_load(self, pt_id: int, load: np.ndarray) -> None:
        """Add a fixed load at a specific point.

        Args:
            pt_id: The ID of the point at which to add the load.
            load: A 2d vector load to apply at the point.

        """
        self.get_point(pt_id).add_fixed_load(load)

    def fix_points(self, pt_id_list: Iterable[int]) -> None:
        """Fix the position of a list of points in the mesh."""
        for pt in [self.get_point(pt_id) for pt_id in pt_id_list]:
            pt.set_fixed_dof("x", "y")

    def fix_all_points(self) -> None:
        """Fix the position of all points in the mesh."""
        self.fix_points([p.id for p in self.points if p.id is not None])

    def rotate_points(self, base_pt_id: int, angle: float, pt_id_list: Iterable[int]) -> None:
        """Rotate all points in a list by a given angle about a base point, counter-clockwise."""
        base_pt = self.get_point(base_pt_id)
        for pt in [self.get_point(pt_id) for pt_id in pt_id_list]:
            pt.rotate(base_pt, angle)

    def rotate_all_points(self, base_pt_id: int, angle: float) -> None:
        """Rotate all points in the mesh by a given angle about a base point, counter-clockwise."""
        self.rotate_points(base_pt_id, angle, [p.id for p in self.points if p.id is not None])

    def move_points(self, dx: float, dy: float, pt_id_list: Iterable[int]) -> None:
        """Move (translate) a selection of points in the x & y directions."""
        for pt in [self.get_point(pt_id) for pt_id in pt_id_list]:
            pt.move(dx, dy)

    def move_all_points(self, dx: float, dy: float) -> None:
        """Move (translate) all points in the x & y directions."""
        self.move_points(dx, dy, [p.id for p in self.points if p.id is not None])

    def scale_all_points(self, sf: float, base_pt_id: int = 0) -> None:
        """Scale the coordinates of all points in the mesh by a constant scaling factor, relative to some base point.

        Args:
            sf: The scale factor.
            base_pt_id: The point from which coordinates are taken to be relative. Defaults to the origin.

        """
        base_pt = self.get_point(base_pt_id).position
        for pt in self.points:
            pt.position = (pt.position - base_pt) * sf + base_pt

    def get_diff(self, pt0: int, pt1: int) -> np.ndarray:
        """The vector pointing from one point to another.

        Args:
            pt0: The source point.
            pt1: The destination point.

        Returns:
            A two-dimensional vector pointing from source to destination.

        """
        return self.get_point(pt1).position - self.get_point(pt0).position

    def get_length(self, pt0: int, pt1: int) -> float:
        """The Cartesian distance of the line between two points."""
        return float(np.linalg.norm(self.get_diff(pt0, pt1)))

    def display(self, disp: bool = False) -> None:
        """Display information about the mesh.

        Args:
            disp: If True, print to console. Otherwise, do nothing.

        """
        if not disp:
            return

        for pt in self.points:
            pt.display()

        logger.info("Curve count: {0}".format(len(self.curves)))
        logger.info("Point count: {0}".format(len(self.points)))

    def plot(self, show: bool = False, save: bool = False, *, filename: str = "meshLayout") -> None:
        """Plot the mesh if we elect to show or save it. Otherwise, do nothing.

        Args:
            show: If True, the plot window will open.
            save: If True, the plot will be saved to a file.
            filename: A filename which can be used when saving the mesh image.

        """
        if not (show or save):
            return

        plt.figure()
        plt.axis("equal")
        plt.xlabel(r"$x$")
        plt.ylabel(r"$y$")

        for curve in self.curves:
            curve.plot()

        for point in self.points:
            point.plot()

        lims = plt.gca().get_xlim()
        ext = (lims[1] - lims[0]) * 0.1
        plt.xlim([lims[0] - ext, lims[1] + ext])

        # Process optional arguments and save or show figure
        if save:
            plt.savefig(f"{filename}.eps", format="eps")
        if show:
            plt.show()  # pragma: no cover

    def write(self, mesh_dir: Path | str = Path("mesh")) -> None:
        """Write the mesh to text files."""
        mesh_dir = Path(mesh_dir)
        mesh_dir.mkdir(exist_ok=True)

        x, y = list(zip(*(pt.position for pt in self.points)))
        write_as_list(mesh_dir / "nodes.txt", ["x", x], ["y", y])

        x, y = list(zip(*(pt.is_dof_fixed for pt in self.points)))
        write_as_list(
            mesh_dir / "fixedDOF.txt",
            ["x", x],
            ["y", y],
            header_format=">1",
            data_format=">1",
        )

        x, y = list(zip(*(pt.fixed_load for pt in self.points)))
        write_as_list(
            mesh_dir / "fixedLoad.txt",
            ["x", x],
            ["y", y],
            header_format=">6",
            data_format="6.4e",
        )

        for sm in self.submesh:
            sm.write(mesh_dir)


class Subcomponent:
    """A child component of the mesh used for splitting up different sets of curves.

    The `Subcomponent` will store a set of curves, as well as a set of line segments once the geometry
    is discretized.

    Attributes:
        name: The name of the `Subcomponent`, which must correspond to the name of the associated `Substructure`.
        mesh: A reference to the parent `Mesh` object, to which this `Subcomponent` belongs.
        curves: A list of control `Curve`s that define the `Subcomponent`.
        line_segments: A list of line segments as a result of the curve being discretized.

    """

    def __init__(self, name: str, mesh: Mesh):
        self.name = name
        self.mesh = mesh
        self.curves: list["Curve"] = []
        self.line_segments: list[Curve] = []

    def add_curve(self, pt_id1: int, pt_id2: int, **kwargs: Any) -> "Curve":
        """Add a curve to the submesh.

        Args:
            pt_id1: The starting point ID.
            pt_id2: The end point ID.

        Keyword Args:
            ID: An explicit ID to use for the curve.
            arcLen: The arclength of the curve (if circular).
            radius: The radius of the curve (if circular).
            Nel: The number of line segments to divide the curve into.

        Returns:
            The Curve object.

        """
        curve = Curve(
            self.mesh.get_point(pt_id1),
            self.mesh.get_point(pt_id2),
            mesh=self.mesh,
            id=kwargs.get("ID"),
            arc_length=kwargs.get("arcLen"),
            radius=kwargs.get("radius"),
            num_elements=kwargs.get("Nel", 1),
        )
        self.curves.append(curve)

        points, line_segments = curve.distribute_points()
        self.mesh.points.extend(points)
        self.line_segments.extend(line_segments)

        return curve

    def write(self, mesh_dir: Path) -> None:
        """Write the submesh to file.

        The submesh is stored as a list of Point IDs for each line segment in the submesh.

        """
        if not self.line_segments:
            return  # pragma: no cover

        point_indices = []
        for line in self.line_segments:
            point_indices.append([line.start_point.index, line.end_point.index])

        pt_l, pt_r = list(zip(*point_indices))
        write_as_list(
            mesh_dir / f"elements_{self.name}.txt",
            ["ptL", pt_l],
            ["ptR", pt_r],
            header_format="<4",
            data_format=">4",
        )


class _ShapeBase:
    """An abstract base class for all Shapes."""

    def __init__(self, id: int | None = None, mesh: Mesh | None = None) -> None:
        self.id = id
        self._mesh = mesh

    @abc.abstractmethod
    def plot(self) -> None:
        raise NotImplementedError


class Point(_ShapeBase):
    """A general point in 2D space, used to represent control points and location of FE nodes.

    Attributes:
        position: A 2D array containing the (x, y) coordinates.
        is_dof_fixed: A length-2 list corresponding to True/False of that DOF if fixed.
        fixed_load: A 2D array containing the externally-applied forces in (x, y) directions.

    """

    def __init__(self, id: int | None = None, mesh: Mesh | None = None) -> None:
        super().__init__(id=id, mesh=mesh)
        self.position = np.zeros(2)
        self.is_dof_fixed = [True, True]
        self.fixed_load = np.zeros(2)
        self._is_used = False

    @property
    def index(self) -> int:
        """The index of the point within the mesh point list."""
        if self._mesh is None:
            raise ValueError("Point is not associated with a mesh")
        return self._mesh.points.index(self)

    @property
    def is_used(self) -> bool:
        """If True, the point will be used in the solution and is free to move."""
        return self._is_used

    @is_used.setter
    def is_used(self, value: bool) -> None:
        self._is_used = value
        if value:
            self.set_free_dof("x", "y")
        else:
            self.set_fixed_dof("x", "y")

    @property
    def x_pos(self) -> float:
        """The x-coordinate of the point."""
        return self.position[0]

    @property
    def y_pos(self) -> float:
        """The y-coordinate of the point."""
        return self.position[1]

    def add_fixed_load(self, load: np.ndarray) -> None:
        """Add a fixed load to the point."""
        self.fixed_load += load

    def set_free_dof(self, *args: str) -> None:
        """Set specific degrees of freedom to be free.

        Args:
            *args: Can be "x" and/or "y".

        """
        for arg in args:
            if arg == "x":
                self.is_dof_fixed[0] = False
            if arg == "y":
                self.is_dof_fixed[1] = False

    def set_fixed_dof(self, *args: str) -> None:
        """Set specific degrees of freedom to be fixed.

        Args:
            *args: Can be "x" and/or "y".

        """
        for arg in args:
            if arg == "x":
                self.is_dof_fixed[0] = True
            if arg == "y":
                self.is_dof_fixed[1] = True

    def move(self, dx: float, dy: float) -> None:
        self.position += np.array([dx, dy])

    def get_x_pos(self) -> float:
        # Kept for backwards-compatibility with old meshDicts
        return self.x_pos

    def get_y_pos(self) -> float:
        # Kept for backwards-compatibility with old meshDicts
        return self.y_pos

    def rotate(self, base_pt: Point | int, angle: float) -> None:
        """Rotate this point about another base point.

        Args:
            base_pt: The point around which to rotate.
            angle: The angle to rotate, in degrees, positive counter-clockwise.

        """
        if isinstance(base_pt, int):
            if self._mesh is None:
                raise LookupError(
                    "Only points existing in the mesh with an ID can be rotated by ID."
                )
            base_pt = self._mesh.get_point(base_pt)
        self.position = (
            trig.rotate_vec_2d(self.position - base_pt.position, angle) + base_pt.position
        )

    def display(self) -> None:
        logger.info(
            " ".join(
                [
                    f"{self.__class__.__name__} {self.index}",
                    f"ID = {self.id}, Pos = {self.position}",
                ]
            )
        )

    def plot(self) -> None:
        """Plot the point.

        If it is a control point with an ID, it is a circle with a label. Otherwise, it's a star.

        """
        if self.id is None:
            plt.plot(self.x_pos, self.y_pos, "r*")
        else:
            plt.plot(self.x_pos, self.y_pos, "ro")
            plt.text(self.x_pos, self.y_pos, f" {self.id}")


class Curve(_ShapeBase):
    """A curve connecting two `Point`s.

    The curve is characterized by its curvature, which is zero by default (a straight line).

    Attributes:
        id: A unique ID for the curve.
        start_point: The starting `Point` of the curve.
        end_point: The end `Point` of the curve.
        curvature: The curvature of the curve, i.e. the inverse of the radius. A value of zero is a straight line.
        num_elements: The number of elements to split the curve into during mesh generation.

    """

    def __init__(
        self,
        start_point: Point,
        end_point: Point,
        *,
        id: int | None = None,
        mesh: Mesh | None = None,
        curvature: float | None = None,
        arc_length: float | None = None,
        radius: float | None = None,
        num_elements: int = 1,
    ):
        super().__init__(id=id, mesh=mesh)
        self.start_point = start_point
        self.end_point = end_point
        self.num_elements = num_elements

        if sum(item is not None for item in (curvature, arc_length, radius)) > 1:
            raise ValueError("Can only set one of 'curvature', 'arc_length', or 'radius'")

        self.curvature = 0.0
        if curvature is not None:
            self.curvature = curvature
        elif arc_length is not None:
            self.arc_length = arc_length
        elif radius is not None:
            self.radius = radius

    @property
    def chord(self) -> float:
        """The chord length, i.e. the distance between start and end points."""
        return float(np.linalg.norm(self.end_point.position - self.start_point.position))

    @property
    def radius(self) -> float:
        """The radius of the curve. A flat curve has an infinite radius."""
        return 1 / self.curvature if self.curvature != 0 else np.inf

    @radius.setter
    def radius(self, value: float) -> None:
        self.curvature = 1 / value if ~np.isinf(value) else 0.0

    def _get_arc_length_residual(self, curvature: float, arc_length: float) -> float:
        """Return the trigonometric residual for the relationship between curvature and arclength."""
        return 0.5 * curvature * self.chord - np.sin(0.5 * curvature * arc_length)

    @property
    def arc_length(self) -> float:
        """The arclength of the curve."""
        if self.curvature == 0:
            return self.chord
        else:

            return fzero(
                lambda s: self._get_arc_length_residual(curvature=self.curvature, arc_length=s),
                self.chord + 1e-6,
            )

    @arc_length.setter
    def arc_length(self, value: float) -> None:
        if self.chord >= value:
            self.curvature = 0.0
        else:
            # Keep increasing guess until fsolve finds the first non-zero root.
            # The criterion here is 1e-6, because the zero case is already handled above
            # and if we set kap = 0, then the residual is always zero, but that case is
            # singular.
            kap0 = 0.02
            while (
                kap := fzero(
                    lambda x: self._get_arc_length_residual(curvature=x, arc_length=value), kap0
                )
            ) <= 1e-6:
                kap0 += 0.02

            self.curvature = kap

    def get_coords(self, s: float) -> np.ndarray:
        """Get the coordinates of any point along the curve.

        Args:
            s: The position along the curve (as a fraction of total arclength).

        Returns:
            The (x,y) coordinates of the point, as a numpy array.

        """
        if self.curvature == 0.0:
            return self.start_point.position * (1 - s) + self.end_point.position * s
        else:
            diff = self.end_point.position - self.start_point.position
            gam = np.arctan2(diff[1], diff[0])
            alf = self.arc_length / (2 * self.radius)
            return (
                self.start_point.position
                + 2.0 * self.radius * np.sin(s * alf) * trig.ang2vec(gam + (s - 1.0) * alf)[:2]
            )

    def distribute_points(self) -> tuple[list[Point], list[Curve]]:
        """Distribute points along the curve during mesh generation.

        Returns:
             A list of `Point`s and straight `Curve` objects connecting them.

        """
        points = [self.start_point]
        if self.num_elements > 1:
            # Distribute N points along a parametric curve defined by f(s), s in [0,1]
            for s in np.linspace(0.0, 1.0, self.num_elements + 1)[1:-1]:
                point = Point(mesh=self._mesh)
                point.is_used = True
                point.position = self.get_coords(s)
                points.append(point)
        points.append(self.end_point)

        lines = []
        for ptSt, ptEnd in zip(points[:-1], points[1:]):
            line = Curve(ptSt, ptEnd, mesh=self._mesh)
            lines.append(line)
        return points, lines

    def plot(self) -> None:
        """Plot the curve as a line by chaining all component points together."""
        points, _ = self.distribute_points()
        x, y = list(zip(*(pt.position for pt in points)))
        plt.plot(x, y, "k-")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['planingfsi', 'planingfsi.fe', 'planingfsi.potentialflow']

package_data = \
{'': ['*']}

install_requires = \
['click-log>=0.4.0,<0.5.0',
 'click>=8.0,<9.0',
 'matplotlib>=3.0,<4.0',
 'numpy>=1.15,<2.0',
 'scipy>=1.8.0,<2.0.0']

entry_points = \
{'console_scripts': ['planingfsi = planingfsi.cli:cli']}

setup_kwargs = {
    'name': 'planingfsi',
    'version': '0.3.1',
    'description': 'Fluid-Structure Interaction for large deformation planing surfaces',
    'long_description': '# PlaningFSI\n\n[![Run Python Tests](https://github.com/mattkram/planingfsi/workflows/Run%20Python%20Tests/badge.svg)](https://github.com/mattkram/planingfsi/actions)\n[![Coverage](https://codecov.io/gh/mattkram/planingfsi/branch/develop/graph/badge.svg)](https://codecov.io/gh/mattkram/planingfsi)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Docs](https://readthedocs.org/projects/planingfsi/badge/?version=latest)](https://planingfsi.readthedocs.io/en/latest/?badge=latest)\n[![Version](https://img.shields.io/pypi/v/planingfsi.svg)](https://pypi.org/project/planingfsi/)\n[![License](https://img.shields.io/pypi/l/planingfsi.svg)](https://pypi.org/project/planingfsi/)\n\nPlaningFSI is a scientific Python program use to calculate the steady-state response of two-dimensional marine structures planing at constant speed on the free surface with consideration for Fluid-Structure Interaction (FSI) and rigid body motion.\nIt was originally written in 2012-2013 to support my Ph.D. research and has recently (2018) been updated and released as open-source.\n\n## Cautionary Note\n\nI am currently working on releasing this package as open source.\nSince this is my first open-source release, the next few releases on PyPI should not be used for production.\nI will release version 1.0.0 and remove this note once I feel that I have sufficiently cleaned up and documented the code.\n\n## Required Python version\n\nThe code is written in Python and was originally written in Python 2.6.5.\nit has since been updated to require Python 3.6+.\n\n## Installation\n\nPlaningFSI can be installed with pip:\n\n```\npip install planingfsi\n```\n\n## Contributing\n\nTo contribute, you should install the code in developer mode.\n\n```\npoetry install --develop=.\n```\n\n## Getting Started\n\nThe main command-line interface is called `planingFSI` and can be called directly, once appropriate input files have been prepared.\nA collection of examples can be found in the `tutorials` directory in the source package.\n',
    'author': 'Matt Kramer',
    'author_email': 'matthew.robert.kramer@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mattkram/planingfsi',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from planingfsi import trig

if TYPE_CHECKING:
    from planingfsi.config import Config
    from planingfsi.fe.rigid_body import RigidBody
    from planingfsi.fe.structure import StructuralSolver
    from planingfsi.fe.substructure import Substructure
    from planingfsi.potentialflow.solver import PotentialPlaningSolver
    from planingfsi.simulation import Simulation


class Figure:
    """A wrapper around a Matplotlib figure to plot the results/status of a PlaningFSI simulation."""

    def __init__(self, simulation: Simulation, *, figsize=(16, 9)):
        self.simulation = simulation

        self.figure = plt.figure(figsize=figsize)
        if self.config.plotting.watch:
            plt.ion()

        self.subplots: list[Subplot] = list(self.create_subplots())

    @property
    def config(self) -> Config:
        return self.simulation.config

    @property
    def solid(self) -> StructuralSolver:
        return self.simulation.structural_solver

    def create_subplots(self) -> Iterator[Subplot]:
        """A generator method that is used to create all the subplots.

        This method can be overridden in a subclass to create different layouts.

        """
        yield GeometrySubplot((0.05, 0.6, 0.9, 0.35), parent=self)
        yield ResidualSubplot((0.40, 0.05, 0.25, 0.45), parent=self)

        if self.solid.rigid_bodies:
            body = self.solid.rigid_bodies[0]
            yield ForceSubplot((0.70, 0.30, 0.25, 0.2), body=body, parent=self)
            yield MotionSubplot((0.70, 0.05, 0.25, 0.2), body=body, parent=self)

        if len(self.solid.rigid_bodies) > 1:
            body = self.solid.rigid_bodies[1]
            yield ForceSubplot((0.05, 0.30, 0.25, 0.2), body=body, parent=self)
            yield MotionSubplot((0.05, 0.05, 0.25, 0.2), body=body, parent=self)

    def update(self) -> None:
        """Update all subplots and redraw. If configured, also save figure as an image."""
        for s in self.subplots:
            s.update(is_final=self.simulation.is_converged)

        plt.draw()

        if self.config.plotting.save:
            self.save()

    def write_time_histories(self) -> None:
        for s in self.subplots:
            if isinstance(s, TimeHistorySubplot):
                s.write()

    def save(self) -> None:
        """Save the figure to a file, numbered by the iteration.

        If it is the first iteration, clear out any existing figures.

        """
        self.simulation.fig_dir.mkdir(exist_ok=True)

        if self.simulation.it == 0:
            for f in self.simulation.fig_dir.glob("*"):
                f.unlink()

        file_path = Path(
            self.simulation.fig_dir,
            f"frame{self.simulation.it:04d}.{self.config.plotting.fig_format}",
        )
        self.figure.savefig(file_path, format=self.config.plotting.fig_format)

    @staticmethod
    def show() -> None:
        try:
            plt.show(block=True)
        except Exception:
            pass


class Subplot:
    """Base class for all subplots."""

    def update(self, is_final: bool = False) -> None:
        raise NotImplementedError


class GeometrySubplot(Subplot):
    """A subplot containing the geometry of the body as well as the free surface
    and pressure profiles.

    """

    def __init__(self, pos: tuple[float, float, float, float], *, parent: Figure):
        self._parent = parent
        self._ax = parent.figure.add_axes(pos)

        self._init_handles()
        self._init_axes()

    def _init_handles(self):
        """Initialize all line handles."""
        (self._handle_nodes,) = self._ax.plot([], [], "ko")
        (self._handle_fs,) = self._ax.plot([], [], "b-")
        (self._handle_fs_init,) = self._ax.plot([], [], "b--")

        # Line handles for element initial and current positions
        self._handles_elements_init = {}
        self._handles_elements = {}

        # Line handles for the air and fluid pressure profiles
        self._handles_air_pressure = {}
        self._handles_hydro_pressure = {}
        for struct in self._parent.solid.substructures:
            (self._handles_air_pressure[struct],) = self._ax.plot([], [], "g-")
            (self._handles_hydro_pressure[struct],) = self._ax.plot([], [], "r-")
            for el in struct.elements:
                (self._handles_elements_init[el],) = self._ax.plot([], [], "k--")
                (self._handles_elements[el],) = self._ax.plot([], [], "k-", linewidth=2)

        self.lineCofR: list[CofRPlot] = []
        for body in self._parent.solid.rigid_bodies:
            # Initial position
            CofRPlot(
                self._ax,
                body,
                grid_len=self._parent.config.plotting.CofR_grid_len,
                cr_style="k--",
                cg_marker="ks",
                fill=False,
            )
            # Current position (will get updated)
            self.lineCofR.append(
                CofRPlot(
                    self._ax,
                    body,
                    grid_len=self._parent.config.plotting.CofR_grid_len,
                    cr_style="k-",
                    cg_marker="ko",
                )
            )

        self._handle_status_text = self._ax.text(
            0.05, 0.95, "", ha="left", va="top", transform=self._ax.transAxes
        )

    def _init_axes(self) -> None:
        """Initialize the axes with the correct limits and scaling."""
        # Find max/min of all nodal coordinates
        x = [nd.x for nd in self.solid.nodes]
        y = [nd.y for nd in self.solid.nodes]
        x_min, x_max = min(x), max(x)
        y_min, y_max = min(y), max(y)

        # Handle explicit setting of max/min dimensions
        if self.config.plotting.xmin is not None:
            x_min = self.config.plotting.xmin
            self.config.plotting.ext_w = 0.0
        if self.config.plotting.xmax is not None:
            x_max = self.config.plotting.xmax
            self.config.plotting.ext_e = 0.0
        if self.config.plotting.ymin is not None:
            y_min = self.config.plotting.ymin
            self.config.plotting.ext_s = 0.0
        if self.config.plotting.ymax is not None:
            y_max = self.config.plotting.ymax
            self.config.plotting.ext_n = 0.0

        # Set the limits & axis labels
        self._ax.set_xlim(
            x_min - (x_max - x_min) * self.config.plotting.ext_w,
            x_max + (x_max - x_min) * self.config.plotting.ext_e,
        )
        self._ax.set_ylim(
            y_min - (y_max - y_min) * self.config.plotting.ext_s,
            y_max + (y_max - y_min) * self.config.plotting.ext_n,
        )
        self._ax.set_aspect("equal")

        self._ax.set_xlabel(r"$x$ [m]")
        self._ax.set_ylabel(r"$y$ [m]")

    @property
    def simulation(self) -> Simulation:
        return self._parent.simulation

    @property
    def solid(self) -> StructuralSolver:
        return self.simulation.structural_solver

    @property
    def fluid(self) -> PotentialPlaningSolver:
        return self.simulation.fluid_solver

    @property
    def config(self) -> Config:
        return self._parent.config

    def _draw_free_surface(self) -> None:
        """Draw the actual and undisturbed free-surface lines."""
        self._handle_fs.set_data(self.fluid.x_coord_fs, self.fluid.z_coord_fs)
        end_pts = np.array([self.config.plotting.x_fs_min, self.config.plotting.x_fs_max])
        self._handle_fs_init.set_data(
            end_pts, self.config.flow.waterline_height * np.ones_like(end_pts)
        )

    @staticmethod
    def _get_pressure_plot_points(ss: Substructure, s0: np.ndarray, p0: np.ndarray) -> np.ndarray:
        """Get coordinates required to plot pressure profile as lines.

        Args:
            ss: The substructure, from which to load coordinates and normal vectors.
            s0: An array of arclengths.
            p0: An array of pressures at the arclength points.

        Returns:
            An array containing x & y coordinates.

        """
        ind = np.abs(p0) > 1e-4
        s0 = s0[ind]
        p0 = p0[ind]

        if s0.size == 0:
            return np.empty((2, 0))

        normal_vec = np.array([ss.get_normal_vector(s) for s in s0])
        coords0 = np.array([ss.get_coordinates(s) for s in s0])
        coords1 = coords0 + ss.config.plotting.pressure_scale * p0[:, np.newaxis] * normal_vec

        # We start with an array of NaN, and then slice in the start & end points for each line
        data = np.full((s0.size * 3, 2), fill_value=np.nan)
        data[::3, :] = coords0
        data[1::3, :] = coords1
        return data.T

    def _draw_pressure_profiles(self, ss: Substructure) -> None:
        """Plot the internal and external pressure profiles as lines."""
        if handle := self._handles_hydro_pressure.get(ss):
            handle.set_data(self._get_pressure_plot_points(ss, ss.s_hydro, ss.p_hydro))
        if handle := self._handles_air_pressure.get(ss):
            handle.set_data(self._get_pressure_plot_points(ss, ss.s_air, -ss.p_air))

    def _draw_substructure(self, ss: Substructure) -> None:
        """Plot the substructure elements and pressure profiles."""
        for el in ss.elements:
            if handle := self._handles_elements.get(el):
                coords = np.array([nd.coordinates for nd in el.nodes])
                handle.set_data(coords.T)

            if handle := self._handles_elements_init.get(el):
                base_pt = np.array([el.parent.rigid_body.x_cr_init, el.parent.rigid_body.y_cr_init])
                pos = np.array(
                    [
                        trig.rotate_point(pos, base_pt, el.parent.rigid_body.trim)
                        - np.array([0, el.parent.rigid_body.draft])
                        for pos in el._initial_coordinates
                    ]
                )
                handle.set_data(pos.T)

        self._draw_pressure_profiles(ss)

    def _draw_structures(self):
        """Plot the structure."""
        for body in self.solid.rigid_bodies:
            for struct in body.substructures:
                self._draw_substructure(struct)

    def update(self, is_final: bool = False) -> None:
        self._draw_structures()
        self._draw_free_surface()
        self._handle_status_text.set_text(
            "\n".join(
                [
                    f"Iteration {self.simulation.it}",
                    f"$Fr={self.config.flow.froude_num:>8.3f}$",
                    f"$\\bar{{P_c}}={self.config.body.PcBar:>8.3f}$",
                ]
            )
        )
        for line in self.lineCofR:
            line.update()


class Series:
    """A container to represent a series on a figure.

    Args:
        x_func: A callable which returns x-coordinates of the series points.
        y_func: A callable which returns y-coordinates of the series points.
        include_history: If True, the full history will be stored and drawn. Otherwise,
            only the latest point will be drawn.
        ax: An axis handle. Defaults to `plt.gca()`.
        label: An optional label to use for the plot legend.
        ignore_first: If True, ignore the first point of the series.
        style: The style to use for the history line.
        style_current: The style to use for the symbol representing the current value.

    """

    def __init__(
        self,
        x_func: Callable[[], float],
        y_func: Callable[[], float],
        *,
        include_history: bool = True,
        ax: Axes | None = None,
        label: str | None = None,
        ignore_first: bool = False,
        style: str = "k-",
        style_current: str = "ro",
    ) -> None:
        self._points: list[tuple[float, float]] = []

        self._get_x = x_func
        self._get_y = y_func

        self._ignore_first = ignore_first

        ax = ax or plt.gca()
        (self.line_handle,) = ax.plot([], [], style, label=label)

        self._current_value_series: Series | None = None
        if include_history:
            self._current_value_series = Series(
                x_func,
                y_func,
                ax=ax,
                include_history=False,
                style=style_current,
            )

    def update(self, is_final: bool = False) -> None:
        """Update the line data by calling the assigned callback functions.

        Args:
            is_final: If True, it is the final iteration and the current value will
                be drawn as a yellow star.
        """
        if self._current_value_series is None:
            self._points = [(self._get_x(), self._get_y())]
        elif self._ignore_first and not self._points:
            self._points.append((np.nan, np.nan))
        else:
            self._points.append((self._get_x(), self._get_y()))

        self.line_handle.set_data(*zip(*self._points))
        if self._current_value_series is not None:
            self._current_value_series.update(is_final)
        elif is_final:
            self.line_handle.set_marker("*")
            self.line_handle.set_markerfacecolor("y")
            self.line_handle.set_markersize(10)


class TimeHistorySubplot(Subplot):
    """A collection of axes (subplot) for displaying iteration series'.

    Args:
        pos: The relative position of the axes in the figure.
        parent: A reference to the parent FSIFigure.
        name: A name for the subplot, used when exporting plot data to file.

    """

    _is_ylog = False

    def __init__(
        self, pos: tuple[float, float, float, float], *, name: str = "default", parent: Figure
    ) -> None:
        self._parent = parent
        self._name = name
        self._ax: list[Axes] = []
        self._add_axes(parent.figure.add_axes(pos))
        self._series = list(self.create_series())

    def _add_axes(self, ax: Axes) -> None:
        """Add a set of child axes."""
        self._ax.append(ax)

    def _add_y_axes(self) -> None:
        """Add a twin y-axis, which shares an x-axis but has a different y-axis scale."""
        self._add_axes(plt.twinx(self._ax[0]))

    def create_series(self) -> Iterator[Series]:
        """A generator that must yield at least one series to plot on the axes."""
        raise NotImplementedError

    def add_series(self, series: Series) -> Series:
        """Add a series to the subplot."""
        self._series.append(series)
        return series

    def set_properties(self, ax_ind: int = 0, **kwargs: Any) -> None:
        """Set the properties of a specific axis by calling matplotlib's `plt.setp()` function.

        Args:
            ax_ind: The index of the axes.
            kwargs: Keyword arguments to pass through to `plt.setp()`.

        """
        plt.setp(self._ax[ax_ind], **kwargs)

    def create_legend(self, ax_ind: int = 0) -> None:
        """Create a legend for a specific set of axes.

        Args:
            ax_ind: The index of the axes.

        """
        self._ax[ax_ind].legend(loc="lower left")

    def update(self, is_final: bool = False) -> None:
        """Update the figure."""
        for s in self._series:
            s.update(is_final)

        for ax in self._ax:
            self._reset_axis_limits(ax)

    def _reset_axis_limits(self, ax: Axes) -> None:
        x_min, x_max = 0, self._parent.simulation.it + 5
        y_min, y_max = np.nan, np.nan

        for series in self._series:
            point_data = np.array(series._points)
            x, y = point_data[~np.isnan(point_data[:, 1])].T

            if x.size == 0:
                continue

            x_min = min(x_min, np.min(x))
            x_max = max(x_max, np.max(x))

            y_min_tmp = np.min(y)
            y_max_tmp = np.max(y)

            # If the y-scale is log, we can only have positive limits
            # And we need to handle the case where y_min/max begins at Nan,
            # where min/max functions don't work
            if not self._is_ylog or y_min_tmp > 0:
                if np.isnan(y_min):
                    y_min = y_min_tmp
                else:
                    y_min = min(y_min, y_min_tmp)

            if not self._is_ylog or y_max_tmp > 0:
                if np.isnan(y_max):
                    y_max = y_max_tmp
                else:
                    y_max = max(y_max, y_max_tmp)

        # This prevents a warning for the first iteration where y_min == y_max
        if y_min == y_max:
            y_min, y_max = y_max - 0.1, y_max + 0.1

        ax.set_xlim(x_min, x_max)
        if ~np.isnan(y_min) and ~np.isnan(y_max):
            ax.set_ylim(y_min, y_max)

    def write(self) -> None:
        """Write the time series results from the axes to a file."""
        with Path(f"{self._name}_timeHistories.txt").open("w") as ff:
            x = [pt[0] for pt in self._series[0]._points]
            for i, x_i in enumerate(x):
                ff.write("{0:4.0f}".format(x_i))
                for s in self._series:
                    ff.write(" {0:8.6e}".format(s._points[i][1]))
                ff.write("\n")


class MotionSubplot(TimeHistorySubplot):
    def __init__(self, pos: tuple[float, float, float, float], *, body: RigidBody, parent: Figure):
        super().__init__(pos, name=body.name, parent=parent)
        self._body = body
        self.set_properties(
            title=r"Motion History: {0}".format(body.name),
            xlabel=r"Iteration",
            ylabel=r"$d$ [m]",
        )
        self.set_properties(1, ylabel=r"$\theta$ [deg]")
        self.create_legend()

    def create_series(self) -> Iterator[Series]:
        yield Series(
            lambda: self._parent.simulation.it,
            lambda: self._body.draft,
            ax=self._ax[0],
            style="b-",
            label="Draft",
        )

        self._add_y_axes()
        yield Series(
            lambda: self._parent.simulation.it,
            lambda: self._body.trim,
            ax=self._ax[1],
            style="r-",
            label="Trim",
        )


class ForceSubplot(TimeHistorySubplot):
    def __init__(self, pos: tuple[float, float, float, float], *, body: RigidBody, parent: Figure):
        super().__init__(pos, name=body.name, parent=parent)
        self._body = body
        self.set_properties(
            title=f"Force & Moment History: {body.name}",
            ylabel=r"$\mathcal{D}/W$, $\mathcal{L}/W$, $\mathcal{M}/WL_c$",
        )
        self.create_legend()

    def create_series(self) -> Iterator[Series]:

        yield Series(
            lambda: self._parent.simulation.it,
            lambda: self._body.loads.L / self._body.weight,
            style="r-",
            label="Lift",
            ignore_first=True,
        )
        yield Series(
            lambda: self._parent.simulation.it,
            lambda: self._body.loads.D / self._body.weight,
            style="b-",
            label="Drag",
            ignore_first=True,
        )
        yield Series(
            lambda: self._parent.simulation.it,
            lambda: self._body.loads.M
            / (self._body.weight * self._parent.config.body.reference_length),
            style="g-",
            label="Moment",
            ignore_first=True,
        )


class ResidualSubplot(TimeHistorySubplot):
    _is_ylog = True

    def __init__(self, pos: tuple[float, float, float, float], *, parent: Figure):
        super().__init__(pos, name="residuals", parent=parent)
        self.set_properties(title="Residual History", xlabel="Iteration", yscale="log")
        self.create_legend()

    def create_series(self) -> Iterator[Series]:
        col = ["r", "b", "g"]
        for body, col_i in zip(self._parent.solid.rigid_bodies, col):
            yield Series(
                lambda: self._parent.simulation.it,
                body.get_res_lift,
                style=f"{col_i}-",
                label=f"Lift: {body.name}",
                ignore_first=True,
            )
            yield Series(
                lambda: self._parent.simulation.it,
                body.get_res_moment,
                style=f"{col_i}--",
                label=f"Moment: {body.name}",
                ignore_first=True,
            )

        yield Series(
            lambda: self._parent.simulation.it,
            lambda: np.abs(self._parent.solid.residual),
            style="k-",
            label="Total",
            ignore_first=True,
        )


class CofRPlot:
    """A set of lines and symbols to represent a body center of gravity and rotation.

    The center of rotation is drawn as two intersecting lines, center of gravity is a black marker.

    Args:
        ax: The axes to which this belongs.
        cr_style: The style to use for the lines.
        grid_len: The length to use for the grid lines.
        cg_marker: The center of gravity marker shape.
        fill: If True, fill the symbols with black.

    """

    def __init__(
        self,
        ax: Axes,
        body: RigidBody,
        *,
        cr_style: str = "k-",
        fill: bool = True,
        cg_marker: str = "o",
        grid_len: float = 0.5,
    ):
        self._ax = ax
        self._body = body
        self._grid_len = grid_len

        (self._cr_handle,) = self._ax.plot([], [], cr_style)
        (self._cg_handle,) = self._ax.plot(
            [], [], cg_marker, markersize=8, markerfacecolor="w" if not fill else "k"
        )
        self.update()

    def update(self) -> None:
        """Update the position and orientation of the handles."""
        c = np.array([self._body.x_cr, self._body.y_cr])
        hvec = trig.rotate_vec_2d(np.array([0.5 * self._grid_len, 0.0]), self._body.trim)
        vvec = trig.rotate_vec_2d(np.array([0.0, 0.5 * self._grid_len]), self._body.trim)
        pts = np.array([c - hvec, c + hvec, c, c - vvec, c + vvec])
        self._cr_handle.set_data(pts.T)
        self._cg_handle.set_data(self._body.x_cg, self._body.y_cg)


def plot_pressure(solver: PotentialPlaningSolver, fig_format: str = "png") -> None:
    """Create a plot of the pressure and shear stress profiles."""
    fig, ax = plt.subplots(1, 1, figsize=(5.0, 5.0))

    for el in solver.pressure_elements:
        ax.plot(*el.plot_coords, color=el.plot_color, linestyle="-")
        ax.plot(el.x_coord * np.ones(2), [0.0, el.pressure], color=el.plot_color, linestyle="--")

    ax.plot(solver.x_coord, solver.pressure, "k-")

    ax.set_xlabel(r"$x\,\mathrm{[m]}$")
    ax.set_ylabel(r"$p\,\mathrm{[kPa]}$")
    ax.set_ylim(ymin=0.0)

    fig.savefig(f"pressureElements.{fig_format}", format=fig_format)



"""Fundamental module for constructing and solving planing potential flow problems."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from scipy.optimize import fmin

from planingfsi import figure
from planingfsi import logger
from planingfsi import solver
from planingfsi import writers
from planingfsi.config import Config
from planingfsi.potentialflow.pressureelement import PressureElement
from planingfsi.potentialflow.pressurepatch import PlaningSurface
from planingfsi.potentialflow.pressurepatch import PressureCushion

if TYPE_CHECKING:
    from planingfsi.potentialflow.pressurepatch import PressurePatch
    from planingfsi.simulation import Simulation


class PotentialPlaningSolver:
    """The top-level object which handles calculation of the potential flow problem.

    Pointers to the planing surfaces and pressure elements are stored for
    reference during initialization and problem setup. The potential planing
    calculation is then solved during iteration with the structural solver.

    Attributes:
        simulation: A reference to the parent Simulation object.
        planing_surfaces: A list of all `PlaningSurface` instances in the domain.
        pressure_cushions: A list of all `PressureCushion` instances in the domain.
        x_coord: An array of x-coordinates of all solution points.
        pressure: An array of pressure at all solution points.
        shear_stress: An array of shear stress at all solution points.
        x_coord_fs: An array of x-coordinates along the free surface.
        z_coord_fs: An array of z-coordinates along the free surface.

    """

    def __init__(self, simulation: Simulation):
        self.simulation = simulation

        self.planing_surfaces: list[PlaningSurface] = []
        self.pressure_cushions: list[PressureCushion] = []

        self.x_coord = np.array([])
        self.pressure = np.array([])
        self.shear_stress = np.array([])
        self.x_coord_fs = np.array([])
        self.z_coord_fs = np.array([])

        self._solver: solver.RootFinder | None = None
        self._wetted_length_it = 0

        self._min_len = np.array([])
        self._max_len = np.array([])
        self._init_len = np.array([])

    @property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        return self.simulation.config

    @property
    def drag_wave(self) -> float:
        """The wave drag, calculated from downstream wave amplitude."""
        f = self.get_free_surface_height
        x_init = -10 * self.config.flow.lam
        (x_trough,) = fmin(f, x_init, disp=False)
        (x_crest,) = fmin(lambda x: -f(x), x_init, disp=False)
        return (
            0.0625
            * self.config.flow.density
            * self.config.flow.gravity
            * (f(x_crest) - f(x_trough)) ** 2
        )

    def __getattr__(self, item: str) -> Any:
        """Calculate total forces as sum of forces on each patch."""
        if item in [
            "drag_total",
            "drag_pressure",
            "drag_friction",
            "lift_total",
            "lift_pressure",
            "lift_friction",
            "moment_total",
        ]:
            return sum(getattr(p, item) for p in self.pressure_patches)
        raise AttributeError

    @property
    def pressure_patches(self) -> list[PressurePatch]:
        """A list of all `PressurePatch` instances in the domain."""
        return self.planing_surfaces + self.pressure_cushions

    @property
    def pressure_elements(self) -> list[PressureElement]:
        """A list of all `PressureElement` instances contained by all `PressurePatch` instances."""
        return [el for patch in self.pressure_patches for el in patch.pressure_elements]

    @property
    def pressure_shear_file_path(self) -> Path:
        return self.simulation.it_dir / f"pressureAndShear.{self.config.io.data_format}"

    @property
    def free_surface_file_path(self) -> Path:
        return self.simulation.it_dir / f"freeSurface.{self.config.io.data_format}"

    @property
    def forces_file_path(self) -> Path:
        return self.simulation.it_dir / f"forces_total.{self.config.io.data_format}"

    def add_planing_surface(
        self, dict_or_instance: dict[str, Any] | PlaningSurface | None, /
    ) -> PlaningSurface:
        """Add planing surface to the calculation from a dictionary file name.

        Args:
            dict_or_instance: A dictionary of values, or a PlaningSurface instance.

        Returns:
            Instance created from dictionary.

        """
        if isinstance(dict_or_instance, PlaningSurface):
            instance = dict_or_instance
        else:
            instance = PlaningSurface(**(dict_or_instance or {}))
        instance.parent = self
        self.planing_surfaces.append(instance)
        return instance

    def add_pressure_cushion(
        self,
        dict_or_instance: dict[str, Any] | PressureCushion | None,
    ) -> PressureCushion:
        """Add pressure cushion to the calculation from a dictionary file name.

        Args:
            dict_or_instance: A dictionary of values, or a PressureCushion instance.

        Returns:
            Instance created from dictionary.

        """
        if isinstance(dict_or_instance, PressureCushion):
            instance = dict_or_instance
        else:
            instance = PressureCushion(**(dict_or_instance or {}))
        instance.parent = self
        self.pressure_cushions.append(instance)
        return instance

    def _calculate_pressure(self) -> None:
        """Calculate pressure of each element to satisfy body boundary conditions."""
        # TODO: This calculation may be sped up by using vectorized functions
        # Form lists of unknown and source elements
        unknown_el = [el for el in self.pressure_elements if not el.is_source and el.width > 0.0]
        nan_el = [el for el in self.pressure_elements if el.width <= 0.0]
        source_el = [el for el in self.pressure_elements if el.is_source]
        # Form arrays to build linear system
        # x location of unknown elements
        x_array = np.array([el.x_coord for el in unknown_el])

        # influence coefficient matrix (to be reshaped)
        influence_matrix = np.array(
            [el.get_influence_coefficient(x) for x in x_array for el in unknown_el]
        )

        # height of each unknown element above waterline
        z_array = np.array([el.z_coord for el in unknown_el]) - self.config.flow.waterline_height

        # subtract contribution from source elements
        z_array -= np.array([np.sum([el.get_influence(x) for el in source_el]) for x in x_array])

        # Solve linear system
        if dim := len(unknown_el):
            p = np.linalg.solve(
                np.reshape(influence_matrix, (dim, dim)), np.reshape(z_array, (dim, 1))
            )[:, 0]

            # Apply calculated pressure to the elements
            for pi, el in zip(p, unknown_el):
                el.pressure = pi

        for el in nan_el:
            el.pressure = 0.0

    def _calculate_residual(self, wetted_length: np.ndarray) -> np.ndarray:
        """Return array of residuals of each planing surface to satisfy Kutta condition.

        Args:
            wetted_length: Array of wetted lengths of planing surfaces.

        Returns:
            Array of trailing edge residuals of each planing surface.

        """
        # Set length of each planing surface
        for Lwi, planing_surface in zip(wetted_length, self.planing_surfaces):
            planing_surface.length = Lwi

        # Update bounds of pressure cushions
        for pressure_cushion in self.pressure_cushions:
            pressure_cushion.update_end_pts()

        # Solve for unknown pressures and output residual
        self._calculate_pressure()

        residual = np.array([p.residual for p in self.planing_surfaces])

        logger.info(f"    Wetted length iteration: {self._wetted_length_it}")
        logger.info(f"      Lw:       {_array_to_string(wetted_length)}")
        logger.info(f"      Residual: {_array_to_string(residual)}\n")

        self._wetted_length_it += 1

        return residual

    def _initialize_solver(self) -> None:
        """Initialize the solver before each call."""
        for p in self.planing_surfaces:
            p.initialize_end_pts()

        if self._solver is None:
            self._init_len = np.array([p.initial_length for p in self.planing_surfaces])
            self._min_len = np.array([p.minimum_length for p in self.planing_surfaces])
            self._max_len = np.array([p.maximum_length for p in self.planing_surfaces])
            self._solver = solver.RootFinder(
                self._calculate_residual,
                self._init_len,
                self.config.solver.wetted_length_solver,
                xMin=self._min_len,
                xMax=self._max_len,
                errLim=self.config.solver.wetted_length_tol,
                firstStep=1e-6,
                maxIt=self.config.solver.wetted_length_max_it_0,
                maxJacobianResetStep=self.config.solver.wetted_length_max_jacobian_reset_step,
                relax=self.config.solver.wetted_length_relax,
            )
        else:
            self._solver.max_it = self.config.solver.wetted_length_max_it
            for i, p in enumerate(self.planing_surfaces):
                length = p.length
                if np.isnan(length) or length - self._min_len[i] < 1e-6:
                    self._init_len[i] = p.initial_length
                else:
                    self._init_len[i] = length

        self._solver.reinitialize(self._init_len)
        self._solver.dx_max_decrease = self.config.solver.wetted_length_max_step_pct_dec * (
            self._init_len - self._min_len
        )
        self._solver.dx_max_increase = self.config.solver.wetted_length_max_step_pct_inc * (
            self._init_len - self._min_len
        )

    def calculate_response(self) -> None:
        """Calculate response, including pressure and free-surface profiles.

        Will load results from file if specified. Otherwise, calculate potential flow problem via
        iteration to find wetted length of all planing surfaces to satisfy all trailing edge
        conditions.

        """
        if self.config.io.results_from_file:
            self._load_results()
            return

        if not self.planing_surfaces:
            # Return early if there are no planing surfaces
            return

        # Initialize planing surface lengths and then solve until residual is *zero*
        self._initialize_solver()
        assert self._solver is not None

        if (self._init_len > 0.0).any():
            logger.info("  Solving for wetted length:")
            self._wetted_length_it = 0
            self._solver.solve()

        # Post-process results from current solution
        self._calculate_free_surface_profile()
        self._calculate_pressure_and_shear_profile()

        # Plot the pressure profile if specified
        if self.config.plotting.show_pressure:
            figure.plot_pressure(self)

    def _calculate_pressure_and_shear_profile(self) -> None:
        """Calculate pressure and shear stress profiles over plate surface."""
        # Calculate forces on each patch
        for p in self.pressure_patches:
            p.calculate_forces()

        # Calculate pressure profile
        if self.pressure_patches:
            self.x_coord = np.unique(np.hstack([p.element_coords for p in self.pressure_patches]))
        else:
            self.x_coord = np.array([-1e6, 1e6])
        self.pressure = np.zeros_like(self.x_coord)
        self.shear_stress = np.zeros_like(self.x_coord)

        for el in self.pressure_elements:
            if el.is_on_body:
                ind = el.x_coord == self.x_coord
                self.pressure[ind] += el.pressure
                self.shear_stress[ind] += el.shear_stress

    def _calculate_free_surface_profile(self) -> None:
        """Calculate free surface profile."""
        x_fs_all: list[np.ndarray] = []  # A list of all sets of points

        # All unique points along all planing surfaces
        x_all_planing = np.unique(
            [x for surf in self.planing_surfaces for x in surf.element_coords if surf.length > 0]
        )
        x_fs_all.append(x_all_planing)

        # Upstream from first plate
        x_fs_all.append(
            _grow_points(
                x_all_planing[-2],
                x_all_planing[-1],
                self.config.plotting.x_fs_max,
                self.config.plotting.growth_rate,
            )
        )
        # Downstream from last plate
        x_fs_all.append(
            _grow_points(
                x_all_planing[1],
                x_all_planing[0],
                self.config.plotting.x_fs_min,
                self.config.plotting.growth_rate,
            )
        )

        # Add points from each pressure cushion
        for patch in self.pressure_cushions:
            if patch.neighbor_down is not None:
                pts_down = np.unique(patch.neighbor_down.element_coords)
            else:
                pts_down = np.array([patch._end_pts[0] - 0.01, patch._end_pts[0]])

            if patch.neighbor_up is not None:
                pts_up = np.unique(patch.neighbor_up.element_coords)
            else:
                pts_up = np.array([patch._end_pts[1], patch._end_pts[1] + 0.01])

            x_mid = 0.5 * (pts_down[-1] + pts_up[0])
            x_fs_all.append(
                _grow_points(pts_down[-2], pts_down[-1], x_mid, self.config.plotting.growth_rate)
            )
            x_fs_all.append(
                _grow_points(pts_up[1], pts_up[0], x_mid, self.config.plotting.growth_rate)
            )

        # Sort x locations and calculate surface heights
        self.x_coord_fs = np.unique(np.hstack(x_fs_all))
        self.z_coord_fs = np.array([self.get_free_surface_height(x) for x in self.x_coord_fs])

    def get_free_surface_height(self, x: float) -> float:
        """Return free surface height at a given x-position considering the
        contributions from all pressure patches.

        Args:
            x: x-position at which free surface height shall be returned.

        Returns:
            Free-surface position at input x-position.

        """
        return sum(patch.get_free_surface_height(x) for patch in self.pressure_patches)

    def write_results(self) -> None:
        """Write results to files."""
        self._write_forces()
        self._write_pressure_and_shear()
        self._write_free_surface()

    def _write_forces(self) -> None:
        """Write forces to file."""
        writers.write_as_dict(
            self.forces_file_path,
            ["Drag", self.drag_total],
            ["WaveDrag", self.drag_wave],
            ["PressDrag", self.drag_pressure],
            ["FricDrag", self.drag_friction],
            ["Lift", self.lift_total],
            ["PressLift", self.lift_pressure],
            ["FricLift", self.lift_friction],
            ["Moment", self.moment_total],
        )

        for patch in self.pressure_patches:
            patch.write_forces()

    def _write_pressure_and_shear(self) -> None:
        """Write pressure and shear stress profiles to data file."""
        writers.write_as_list(
            self.pressure_shear_file_path,
            ["x [m]", self.x_coord],
            ["p [Pa]", self.pressure],
            ["shear_stress [Pa]", self.shear_stress],
        )

    def _write_free_surface(self) -> None:
        """Write free-surface profile to file."""
        writers.write_as_list(
            self.free_surface_file_path,
            ["x [m]", self.x_coord_fs],
            ["y [m]", self.z_coord_fs],
        )

    def _load_results(self) -> None:
        """Load results from file."""
        self._load_forces()
        self._load_pressure_and_shear()
        self._load_free_surface()

    def _load_forces(self) -> None:
        """Load forces from file."""
        for patch in self.pressure_patches:
            patch.load_forces()

    def _load_pressure_and_shear(self) -> None:
        """Load pressure and shear stress from file."""
        self.x_coord, self.pressure, self.shear_stress = np.loadtxt(
            self.pressure_shear_file_path, unpack=True
        )
        for el in [el for patch in self.planing_surfaces for el in patch.pressure_elements]:
            compare = np.abs(self.x_coord - el.x_coord) < 1e-6
            if any(compare):
                el.pressure = self.pressure[compare][0]
                el.shear_stress = self.shear_stress[compare][0]

        for p in self.planing_surfaces:
            p.calculate_forces()

    def _load_free_surface(self) -> None:
        """Load free surface coordinates from file."""
        try:
            self.x_coord_fs, self.z_coord_fs = np.loadtxt(self.free_surface_file_path, unpack=True)
        except IOError:
            self.z_coord_fs = np.zeros_like(self.x_coord_fs)


def _array_to_string(array: np.ndarray) -> str:
    """Convert an array to a string."""
    return f'[{", ".join("{0:11.4e}".format(a) for a in array)}]'


def _grow_points(x0: float, x1: float, x_max: float, rate: float = 1.1) -> np.ndarray:
    """Grow points exponentially from two starting points assuming a growth rate.

    Args:
        x0: The first point.
        x1: The second point.
        x_max: The maximum distance.
        rate: The growth rate of spacing between subsequent points.

    """
    # https://math.stackexchange.com/a/1897065
    n = 0
    while (abs(x1 - x0) * (1 - rate**n) / (1 - rate)) < abs(x_max - x0):
        n += 1
    lengths = (x1 - x0) * rate ** np.array(range(n))
    distances = np.cumsum(lengths)
    return x0 + np.hstack((np.array([0]), distances))



"""PlaningFSI is a Python package for solving 2d FSI problems.

PlaningFSI solves the general FSI problem of 2-D planing surfaces on a free
surface, where each surface can be rigid or flexible. The substructures are
part of a rigid body, which can be free in sinkage and trim.

The fluid solver is based on linearized potential-flow assumptions,
and the structural solver considers a large-deformation simple beam element.

"""
import logging

logging.basicConfig()

logger = logging.getLogger("planingfsi")
logger.setLevel(logging.INFO)

from planingfsi.fe.femesh import Mesh
from planingfsi.fe.substructure import FlexibleMembraneSubstructure
from planingfsi.fe.substructure import RigidSubstructure
from planingfsi.fe.substructure import TorsionalSpringSubstructure
from planingfsi.potentialflow.pressurepatch import PlaningSurface
from planingfsi.potentialflow.pressurepatch import PressureCushion
from planingfsi.simulation import Simulation

__all__ = [
    "logger",
    "Mesh",
    "FlexibleMembraneSubstructure",
    "RigidSubstructure",
    "TorsionalSpringSubstructure",
    "PlaningSurface",
    "PressureCushion",
    "Simulation",
]



"""Solvers for single and multi-dimensional nonlinear problems."""
from __future__ import annotations

from collections.abc import Callable
from typing import Any

import numpy


class RootFinder:
    """A generalized root finder for solving nonlinear equations.

    Args:
        func: A callable which returns an array of residuals the same shape as the argument.
        xo: An array with an initial guess at the solution.
        method: The method to use for solving. Options are "broyden" and "secant".

    """

    def __init__(
        self,
        func: Callable[[numpy.ndarray], numpy.ndarray],
        xo: numpy.ndarray,
        method: str,
        **kwargs: Any,
    ):
        self.func = func
        self.x = numpy.array(xo)
        self.f = numpy.zeros_like(xo)
        self.dim = len(xo)

        self.x_min = numpy.array(kwargs.get("xMin", -numpy.ones_like(xo) * float("Inf")))
        self.x_max = numpy.array(kwargs.get("xMax", numpy.ones_like(xo) * float("Inf")))
        self.max_it = kwargs.get("maxIt", 100)
        self.dx_init = kwargs.get("firstStep", 1e-6)
        self.error_limit = kwargs.get("errLim", 1e-6)
        self.relax = kwargs.get("relax", 1.0)

        dx_max = kwargs.get("dxMax", numpy.ones_like(xo) * float("Inf"))
        self.dx_max_increase = numpy.array(kwargs.get("dxMaxInc", dx_max))
        self.dx_max_decrease = numpy.array(kwargs.get("dxMaxDec", dx_max))

        self.derivative_method = kwargs.get("derivativeMethod", "right")
        self.err = 1.0
        self.it = 0
        self.jacobian: numpy.ndarray | None = None
        self.x_prev = numpy.zeros((self.dim,))
        self.f_prev = numpy.zeros((self.dim,))
        self.max_jacobian_reset_step = kwargs.get("maxJacobianResetStep", 5)
        self.is_converged = False

        # Calculate function value at initial point
        self.dx: numpy.ndarray = numpy.zeros_like(xo)
        self.df = numpy.zeros_like(xo)

        self.evaluate_function()

        self.jacobian_step = 0

        # TODO: Replace with factory method and method overloading
        if method.lower() == "broyden":
            self.get_step = self.get_step_broyden
        else:
            self.get_step = self.get_step_secant

    def reinitialize(self, xo: numpy.ndarray) -> None:
        """Re-initialize the solver for a new solution."""
        self.err = 1.0
        self.it = 0
        self.jacobian = None

        # Calculate function value at initial point
        self.x = xo
        self.evaluate_function()

    def limit_step(self, dx: numpy.ndarray = None) -> numpy.ndarray:
        if dx is None:
            dx = self.dx
        assert dx is not None
        dx *= self.relax

        x = self.x + dx

        x = numpy.max(numpy.vstack((x, self.x_min)), axis=0)
        x = numpy.min(numpy.vstack((x, self.x_max)), axis=0)

        dx = x - self.x

        dx_lim_pct = numpy.ones_like(dx)
        for i in range(len(dx_lim_pct)):
            assert dx is not None
            if dx[i] > 0:
                dx_lim_pct[i] = numpy.min([dx[i], self.dx_max_increase[i]]) / dx[i]
            elif dx[i] < 0:
                dx_lim_pct[i] = numpy.max([dx[i], -self.dx_max_decrease[i]]) / dx[i]

        dx *= dx_lim_pct
        assert dx is not None
        self.dx = dx

        return dx

    def store_previous_step(self) -> None:
        """Store the results from the previous step."""
        self.x_prev = self.x * 1.0
        self.f_prev = self.f * 1.0

    def evaluate_error(self) -> None:
        """Evaluate the error residual."""
        self.err = numpy.max(numpy.abs(self.dx + 1e-8))
        if self.df is not None:
            self.err += numpy.max(numpy.abs(self.df + 1e-8))
        else:
            self.err += numpy.max(numpy.abs(self.f))

    def evaluate_function(self) -> None:
        """Evaluate the function and calculated the change."""
        self.f = self.func(self.x)
        if self.f_prev is not None:
            self.df = self.f - self.f_prev
        else:
            self.df = None

    def get_step_secant(self) -> numpy.ndarray:
        """Get the step using the Secant method."""
        if self.it == 0:
            self.dx = numpy.ones_like(self.x) * self.dx_init
        else:
            self.dx = -self.f * (self.x - self.x_prev) / (self.f - self.f_prev + 1e-8)
        return self.dx

    def take_step(self, dx: numpy.ndarray = None) -> None:
        """Take the step in the solution."""
        if dx is not None:
            self.dx = dx
        self.store_previous_step()
        self.x += self.dx
        self.evaluate_function()
        self.it += 1

    def reset_jacobian(self) -> None:
        """Reset the Jacobian."""
        fo = self.f * 1.0
        xo = self.x * 1.0

        self.jacobian = numpy.zeros((self.dim, self.dim))
        for i in range(self.dim):
            self.dx = numpy.zeros_like(self.x)
            self.dx[i] = self.dx_init
            self.take_step()
            self.jacobian[:, i] = self.df / self.dx[i]
            self.f = fo
            self.x = xo

        self.jacobian_step = 0

    def get_step_broyden(self) -> numpy.ndarray:
        """Get the next step using Broyden's method."""
        if self.it == 0 or self.jacobian is None:
            self.reset_jacobian()

        # TODO: Consider removing this and fixing static types
        assert self.jacobian is not None

        dx = numpy.reshape(self.dx, (self.dim, 1))
        df = numpy.reshape(self.df, (self.dim, 1))

        self.jacobian += (
            numpy.dot(df - numpy.dot(self.jacobian, dx), dx.T) / numpy.linalg.norm(dx) ** 2
        )
        dx *= 0.0
        dof = [
            not x <= xMin and not x >= xMax for x, xMin, xMax in zip(self.x, self.x_min, self.x_max)
        ]
        if any(dof):
            assert self.jacobian is not None
            b = self.f.reshape(self.dim, 1)
            dx[numpy.ix_(dof)] = numpy.linalg.solve(
                -self.jacobian[numpy.ix_(dof, dof)], b[numpy.ix_(dof)]
            )

        if any(numpy.abs(self.f) - numpy.abs(self.f_prev) > 0.0):
            self.jacobian_step += 1

        if self.jacobian_step >= self.max_jacobian_reset_step:
            dx = numpy.ones_like(dx) * self.dx_init
            self.jacobian = None

        self.dx = dx.reshape(self.dim)

        return self.dx

    def solve(self) -> numpy.ndarray:
        """Solve the nonlinear problem."""
        while self.err >= self.error_limit and self.it < self.max_it:
            self.get_step()
            self.limit_step()
            self.take_step()
            self.evaluate_error()

        self.is_converged = self.err < self.error_limit and not any(self.x <= self.x_min)

        return self.x


def fzero(
    func: Callable[[float], float],
    x_init: float,
    *,
    max_it: int = 100,
    first_step: float = 1e-6,
    err_lim: float = 1e-6,
) -> float:
    """Find the root of a scalar function func with an initial guess x_init.

    Args:
        func: The function for which to find the zero-crossing point.
        x_init: The initial guess of the function's solution.
        max_it: Maximum number of iterations.
        first_step: Length of first step.
        err_lim: Tolerance for iteration residual.

    Returns:
        The value x such that func(x) == 0.0, within a given tolerance.

    """
    error = 1.0
    it_num = 0
    x_old = x_init
    f_old = func(x_old)
    x_new = x_old + first_step
    while error >= err_lim and it_num < max_it:
        f_new = func(x_new)
        delta_x = -f_new * (x_new - x_old) / (f_new - f_old)
        x_old, f_old = x_new, f_new
        x_new += delta_x
        error = numpy.abs(x_new - x_old)
        it_num += 1
    return x_new
